import random

import torch
import torch.nn.functional as F


def SupConHardLoss(model_emb, temp, n_pos):
    '''
    return the SupCon-Hard loss
    features:
        model output embedding, dimension [bsz, n_all, out_dim],
        where bsz is batchsize,
        n_all is anchor, pos, neg (n_all = 1 + n_pos + n_neg)
        and out_dim is embedding dimension
    temp:
        temperature
    n_pos:
        number of positive examples per anchor
    '''
    # l2 normalize every embedding
    features = F.normalize(model_emb, dim=-1, p=2)
    # features_T is [bsz, outdim, n_all], for performing batch dot product
    features_T = torch.transpose(features, 1, 2)
    # anchor is the first embedding
    anchor = features[:, 0]
    # anchor is the first embedding
    anchor_dot_features = torch.bmm(anchor.unsqueeze(1), features_T) / temp
    # anchor_dot_features now [bsz, n_all], contains
    anchor_dot_features = anchor_dot_features.squeeze(1)
    # deduct by max logits, which will be 1/temp since features are L2 normalized
    logits = anchor_dot_features - 1 / temp
    # the exp(z_i dot z_a) excludes the dot product between itself
    # exp_logits is of size [bsz, n_pos+n_neg]
    exp_logits = torch.exp(logits[:, 1:])
    exp_logits_sum = n_pos * torch.log(exp_logits.sum(1))  # size [bsz], scale by n_pos
    pos_logits_sum = logits[:, 1:n_pos + 1].sum(1)  # sum over all (anchor dot pos)
    log_prob = (pos_logits_sum - exp_logits_sum) / n_pos
    loss = - log_prob.mean()
    return loss



def mine_hard_negative(dist_map, knn=10):
    # print("The number of unique EC numbers: ", len(dist_map.keys()))
    ecs = list(dist_map.keys())
    negative = {}
    for i, target in enumerate(ecs):
        sort_orders = sorted(
            dist_map[target].items(), key=lambda x: x[1], reverse=False)
        if sort_orders[1][1] != 0:
            freq = [1 / i[1] for i in sort_orders[1:1 + knn]]
            neg_ecs = [i[0] for i in sort_orders[1:1 + knn]]
        elif sort_orders[2][1] != 0:
            freq = [1 / i[1] for i in sort_orders[2:2 + knn]]
            neg_ecs = [i[0] for i in sort_orders[2:2 + knn]]
        elif sort_orders[3][1] != 0:
            freq = [1 / i[1] for i in sort_orders[3:3 + knn]]
            neg_ecs = [i[0] for i in sort_orders[3:3 + knn]]
        else:
            freq = [1 / i[1] for i in sort_orders[4:4 + knn]]
            neg_ecs = [i[0] for i in sort_orders[4:4 + knn]]

        normalized_freq = [i / sum(freq) for i in freq]
        negative[target] = {
            'weights': normalized_freq,
            'negative': neg_ecs
        }
    return negative


def mine_negative(anchor, id_ec, ec_id, mine_neg):
    anchor_ec = id_ec[anchor]
    pos_ec = random.choice(anchor_ec)
    neg_ec = mine_neg[pos_ec]['negative']
    weights = mine_neg[pos_ec]['weights']
    result_ec = random.choices(neg_ec, weights=weights, k=1)[0]
    while result_ec in anchor_ec:
        result_ec = random.choices(neg_ec, weights=weights, k=1)[0]
    neg_id = random.choice(ec_id[result_ec])
    return neg_id


def random_positive(id, id_ec, ec_id):
    pos_ec = random.choice(id_ec[id])
    pos = id
    if len(ec_id[pos_ec]) == 1:
        return pos + '_' + str(random.randint(0, 9))
    while pos == id:
        pos = random.choice(ec_id[pos_ec])
    return pos


class MultiPosNegDatasetWithMineEC(torch.utils.data.Dataset):

    def __init__(self, id_ec, ec_id, mine_neg, n_pos, n_neg, train_dataset):
        self.id_ec = id_ec
        self.ec_id = ec_id
        self.n_pos = n_pos
        self.n_neg = n_neg
        self.full_list = []
        self.mine_neg = mine_neg
        for ec in ec_id.keys():
            if '-' not in ec:
                self.full_list.append(ec)

        self.train_dataset = train_dataset

    def __len__(self):
        return len(self.full_list)

    def __getitem__(self, index):
        anchor_ec = self.full_list[index]
        anchor = random.choice(self.ec_id[anchor_ec])
        a = torch.from_numpy(self.train_dataset.features["place_holder"][anchor]).unsqueeze(0)
        data = [a]
        for _ in range(self.n_pos):
            pos = random_positive(anchor, self.id_ec, self.ec_id)
            p = torch.from_numpy(self.train_dataset.features["place_holder"][pos]).unsqueeze(0)
            data.append(p)
        for _ in range(self.n_neg):
            neg = mine_negative(anchor, self.id_ec, self.ec_id, self.mine_neg)
            n = torch.from_numpy(self.train_dataset.features["place_holder"][neg]).unsqueeze(0)
            data.append(n)
        return torch.cat(data)


def get_dataloader(dist_map, id_ec, ec_id, n_pos, n_neg, train_dataset, batch_size):
    params = {
        'batch_size': batch_size,
        'shuffle': True,
    }
    negative = mine_hard_negative(dist_map, 100)
    train_data = MultiPosNegDatasetWithMineEC(
        id_ec, ec_id, negative, n_pos, n_neg, train_dataset)
    train_loader = torch.utils.data.DataLoader(train_data, **params)
    return train_loader





from abc import abstractmethod
from typing import Any, List, Tuple

import numpy as np
from joblib import Parallel, delayed
from numpy import ndarray

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization._utils import call_set_features_names
from plants_sm.transformation._utils import tqdm_joblib
from plants_sm.transformation.transformer import Transformer

from tqdm import tqdm


class FeaturesGenerator(Transformer):

    device: str = "cpu"
    output_shape_dimension: int = 2
    features_names: List[str] = []

    @abstractmethod
    def set_features_names(self):
        """
        Abstract method that has to be implemented by all feature generators to set the features names
        """
        raise NotImplementedError

    def _fit(self, dataset: Dataset, instance_type: str) -> 'FeaturesGenerator':
        """
        Abstract method that has to be implemented by all feature generators

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        """

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'FeaturesGenerator':
        """
        Abstract method that has to be implemented by all feature generators

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        """

    @call_set_features_names
    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        General method that calls _featurize that has to be implemented by all feature generators

        Parameters
        ----------
        dataset: Dataset
            dataset to be transformed where instances are the representation or object to be processed.

        Returns
        -------
        dataset with features: Dataset
            dataset object with features
        """
        instances = dataset.get_instances(instance_type)
        if self.n_jobs > 1:
            parallel_callback = Parallel(n_jobs=self.n_jobs, backend="multiprocessing", prefer="threads")
            with tqdm_joblib(tqdm(desc=self.__class__.__name__, total=len(instances.items()))):
                res = parallel_callback(
                    delayed(self._featurize_and_add_identifier)(instance_representation, instance_id)
                    for instance_id, instance_representation in instances.items())
        else:
            res = []
            pbar = tqdm(desc=self.__class__.__name__, total=len(instances.items()))
            for instance_id, instance_representation in dataset.get_instances(instance_type).items():
                res.append(self._featurize_and_add_identifier(instance_representation, instance_id))
                pbar.update(1)

        dataset.add_features(instance_type, dict(res))
        dataset.features_fields[instance_type] = self.features_names

        return dataset

    def _featurize_and_add_identifier(self, instance: Any, identifier: str) -> Tuple[str, ndarray]:
        """
        Private method that calls the _featurize method and returns the dataframe with the features, adding the instance
        identifier to the dataframe.

        It is used to featurize a single instance.

        Parameters
        ----------
        instance: Any
            instance to be featurized.

        identifier: str
            identifier of the instance.

        Returns
        -------

        """
        try:
            features_values = self._featurize(instance)
        # TODO: catch the correct exception
        except Exception:
            features_values = np.zeros(len(self.features_names))

        temp_feature_dictionary = (identifier, features_values)

        return temp_feature_dictionary

    @abstractmethod
    def _featurize(self, instance: Any) -> np.ndarray:
        """
        Method to be implemented by all feature generators to generate features for one instance at a time

        Parameters
        ----------
        instance: Any
            representation or object to be processed by the feature generator

        Returns
        -------
        np.ndarray
        """



import numpy as np
from transformers import AutoTokenizer, AutoModel

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from plants_sm.featurization.proteins.bio_embeddings._utils import get_device, reduce_per_protein


class ProtBert(FeaturesGenerator):

    embedding_dimension = 1024
    output_shape_dimension: int = 2

    def set_features_names(self):
        """
        The method features_names will set the names of the features.
        """
        self.features_names = [f"prot_bert_{num}" for num in range(1, self.embedding_dimension + 1)]

    def _fit(self, dataset: Dataset, instance_type: str) -> 'ProtBert':
        """
        Fit the ProtBert model to the dataset.

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        self: Estimator
            the fitted ProtBert
        """
        self.tokenizer = AutoTokenizer.from_pretrained("Rostlab/prot_bert", do_lower_case=False)
        model = AutoModel.from_pretrained("Rostlab/prot_bert")

        device_to_use = get_device(self.device)
        self.model = model.to(device_to_use)
        self.model = model.eval()

        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'ProtBert':
        """
        Fit the ProtBert model to the dataset.

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        instance_type: str
            type of the instances to be featurized

        Returns
        -------
        self: Estimator
            the fitted ProtBert
        """
        return self._fit(dataset, instance_type)

    def _featurize(self, sequence: str) -> np.ndarray:
        """
        The method _featurize will generate the desired features for a given protein sequence

        Parameters
        ----------
        sequence: str
            instance to be featurized

        Returns
        -------
        features: np.ndarray
            features generated for the given protein sequence

        """
        sequence = ' '.join([char for char in sequence])
        encoded_input = self.tokenizer(sequence, return_tensors='pt')
        encoded_input.to(get_device(self.device))
        output = self.model(**encoded_input)
        tensor = output['last_hidden_state'].detach().cpu().numpy()
        embedding = tensor[0][1:tensor[0].shape[0] - 1, :]

        if self.output_shape_dimension == 2:
            return reduce_per_protein(embedding)

        elif self.output_shape_dimension == 3:
            return embedding



from functools import wraps


def fit_status(func):

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # TODO: implement try/except block to catch for fitting errors rather than ValueError
        try:
            # it is required to set fitted twice.
            # the first time allows to set the estimated parameters,
            # as these can be accessed without fitting the estimator
            # the second time is to set the status to True,
            # even if some parameters are internally altered during the fitting process due to parameter processing
            self._fitted = True
            res = func(self, *args, **kwargs)
            self._fitted = True
            return res

        except ValueError as e:
            self._fitted = False
            raise e

    return wrapper


from abc import abstractmethod
from itertools import tee
from typing import List, Iterable

from joblib import Parallel, delayed
from statistics import mean, stdev


class Descriptor:
    """
    A descriptor is a function that takes a list of atoms and returns a list of floats.
    """

    @abstractmethod
    def __call__(self, sequence: str, **kwargs):
        """
        It calculates the descriptor for a given sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the descriptor for.

        kwargs
            Keyword arguments that are passed to the descriptor.

        Returns
        -------
        list of float
            A list of floats.
        """
        pass

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__class__.__name__

    @abstractmethod
    def get_features_out(self, **kwargs):
        """
        Returns the features that this descriptor will return.

        Parameters
        ----------
        kwargs
            Keyword arguments that are passed to the descriptor.

        Returns
        -------
        list
            A list of strings.
        """
        pass


def calculate_descriptor(descriptor: Descriptor, sequences: List[str], n_jobs: int = 1, **kwargs):
    """
    It calculates the descriptor for a list of sequences in parallel.

    Parameters
    ----------
    descriptor : Descriptor
        The descriptor to make parallel.

    sequences : list of str
        The sequences to calculate the descriptor for.

    n_jobs : int
        The number of jobs to run in parallel.

    kwargs
        Keyword arguments that are passed to the descriptor.

    Returns
    -------
    list of list of float
        The list of the descriptor for each sequence.
    """
    parallel = Parallel(n_jobs=n_jobs)
    return parallel(delayed(descriptor)(sequence, **kwargs) for sequence in sequences)


def normalize_aap(aap: dict) -> dict:
    """
    Normalize each amino acid property.

    Parameters
    ----------
    aap : dict
        The amino acid property.

    Returns
    -------
    aap_norm : dict
        The normalized amino acid property.
    """
    _mean = mean(list(aap.values()))
    _stdev = stdev(list(aap.values()))

    result = {aa: (j - _mean) / _stdev for aa, j in aap.items()}
    return result


def pairwise(iterable: Iterable, n: int = 1) -> Iterable:
    """
    s -> (s0,s1), (s1,s2), (s2, s3), ..., n = 1
    s -> (s0,s2), (s1,s3), (s2, s4), ..., n = 2

    Parameters
    ----------
    iterable : iterable
        The iterable to pairwise.

    n : int, optional (default = 1)
        The number of elements to pairwise.

    Returns
    -------
    list of tuples iterator : Iterable
        The list of tuples.
    """
    a, b = tee(iterable)
    for i in range(n):
        next(b, None)
    return zip(a, b)



from deepmol.compound_featurization import MorganFingerprint, MACCSkeysFingerprint, LayeredFingerprint, RDKFingerprint, \
    AtomPairFingerprint, TwoDimensionDescriptors

DEEPMOL_PRESETS = {
    "morgan_fingerprints": MorganFingerprint,
    "maccs_keys_fingerprints": MACCSkeysFingerprint,
    "layered_fingerprints": LayeredFingerprint,
    "rdk_fingerprints": RDKFingerprint,
    "atompair_fingerprints": AtomPairFingerprint,
    "2d_descriptors": TwoDimensionDescriptors,
}



import os
from typing import Union, List, Dict

import numpy as np

from plants_sm.data_structures.dataset import Dataset, PLACEHOLDER_FIELD
from plants_sm.io.json import write_json, read_json
from plants_sm.io.pickle import write_pickle, read_pickle
from plants_sm.models.enumerators import ModelFileEnumeratorsUtils
from plants_sm.models.model import Model
from plants_sm.transformation.transformer import Transformer


class Pipeline:

    steps: Dict[str, List[Transformer]]
    models: List[Model]
    metrics: List[callable]

    def __init__(self, steps: Union[List[Transformer], Dict[str, List[Transformer]]] = None,
                 models: List[Model] = None,
                 metrics: List[callable] = None):
        """
        Constructor

        Parameters
        ----------
        steps: List[Transformer] | Dict[str, List[Transformer]]
            list of steps to be executed
        models: List[Model]
            list of models to be executed
        metrics: List[callable]
            list of metrics to be executed
        """

        if isinstance(steps, list):
            self.steps = {PLACEHOLDER_FIELD: steps}
        else:
            self.steps = steps
        self.models = models
        self.metrics = metrics
        self._best_model_name = None

        self._models_indexes = {}
        for model in self.models:
            model_name = model.name
            if model_name in self._models_indexes:
                raise ValueError(f"Model with name {model_name} already exists")
            self._models_indexes[model_name] = model

    def fit(self, train_dataset: Dataset, validation_dataset: Dataset = None) -> 'Pipeline':
        """
        Fit the pipeline

        Parameters
        ----------
        train_dataset: Dataset
            dataset to fit the pipeline
        validation_dataset: Dataset
            dataset to validate the pipeline

        Returns
        -------
        self: Pipeline
            fitted pipeline
        """

        for instance_type in self.steps.keys():
            for step in self.steps[instance_type]:
                step.fit_transform(train_dataset, instance_type=instance_type)
                if validation_dataset is not None:
                    step.transform(validation_dataset, instance_type=instance_type)

        for model in self.models:
            model.fit(train_dataset, validation_dataset)
            self._models_indexes[model.name] = model

        return self

    def _transform_dataset(self, dataset: Dataset) -> None:
        """
        Transform the dataset according to the pipeline

        Parameters
        ----------
        dataset: Dataset
            dataset to transform
        """

        for instance_type in self.steps.keys():
            for step in self.steps[instance_type]:
                step.transform(dataset, instance_type=instance_type)

    def predict(self, test_dataset: Dataset, model_name: str = None) -> np.ndarray:
        """
        Predict the dataset according to the pipeline

        Parameters
        ----------
        test_dataset: Dataset
            dataset to predict
        model_name: str
            name of the model to use

        Returns
        -------
        predictions: np.ndarray
            predictions of the dataset
        """

        self._transform_dataset(test_dataset)

        if model_name is not None:
            return self._models_indexes[model_name].predict(test_dataset)
        elif self._best_model_name is not None:
            return self._models_indexes[self._best_model_name].predict(test_dataset)
        else:
            return self._models_indexes[self.models[0].name].predict(test_dataset)

    def predict_proba(self, test_dataset: Dataset, model_name: str = None) -> np.ndarray:
        """
        Predict the dataset according to the pipeline

        Parameters
        ----------
        test_dataset: Dataset
            dataset to predict
        model_name: str
            name of the model to use

        Returns
        -------
        predictions: np.ndarray
            predictions of the dataset
        """

        self._transform_dataset(test_dataset)

        if model_name is not None:
            return self._models_indexes[model_name].predict_proba(test_dataset)
        elif self._best_model_name is not None:
            return self._models_indexes[self._best_model_name].predict_proba(test_dataset)
        else:
            return self._models_indexes[self.models[0].name].predict_proba(test_dataset)

    def _save(self, path: str) -> None:
        """
        Save the pipeline

        Parameters
        ----------
        path: str
            path to save the pipeline
        """

        config = {
            "steps": {},
            "models": [],
            "metrics": [],
            "_best_model_name": self._best_model_name
        }

        for instance_type in self.steps.keys():

            i = 1
            config["steps"][instance_type] = []

            for step in self.steps[instance_type]:
                if instance_type == PLACEHOLDER_FIELD:

                    step_file_path = f"step_{i}.pkl"
                    config["steps"][PLACEHOLDER_FIELD].append(step_file_path)

                    step.to_pickle(os.path.join(path, step_file_path))
                else:

                    step_file_path = f"step_{instance_type}_{i}.pkl"
                    config["steps"][instance_type].append(step_file_path)

                    step.to_pickle(os.path.join(path, f"step_{instance_type}_{i}.pkl"))
                i += 1

        for model in self.models:
            config["models"].append(model.name)
            model.save(os.path.join(path, model.name))

        if self.metrics is not None:
            for metric in self.metrics:
                config["metrics"].append(f"{metric.__name__}.pkl")
                write_pickle(metric, os.path.join(path, f"{metric.__name__}.pkl"))

        write_json(os.path.join(path, "config.json"), config)

    def save(self, path: str) -> None:
        """
        Save the pipeline

        Parameters
        ----------
        path: str
            path to save the pipeline
        """

        if os.path.exists(path):
            if os.path.isdir(path):
                self._save(path)
            else:
                raise ValueError("The path must be a directory")
        else:
            os.mkdir(path)
            self._save(path)

    @classmethod
    def _load(cls, path: str) -> 'Pipeline':
        """
        Load the pipeline

        Parameters
        ----------
        path: str
            path to load the pipeline

        Returns
        -------
        pipeline: Pipeline
            loaded pipeline
        """

        config = read_json(os.path.join(path, "config.json"))
        steps = {}
        models = []
        metrics = []

        for instance_type in config["steps"].keys():

            if instance_type == PLACEHOLDER_FIELD:
                steps[PLACEHOLDER_FIELD] = []
            else:
                steps[instance_type] = []

            for step in config["steps"][instance_type]:
                if instance_type == PLACEHOLDER_FIELD:
                    steps[PLACEHOLDER_FIELD].append(read_pickle(os.path.join(path, step)))
                else:
                    steps[instance_type].append(read_pickle(os.path.join(path, step)))

        for model in config["models"]:
            model_type = ModelFileEnumeratorsUtils.get_class_to_load_from_directory(os.path.join(path, model))
            loaded_model = model_type.load(os.path.join(path, model))
            models.append(loaded_model)

        for metric_path in config["metrics"]:
            metric = read_pickle(os.path.join(path, metric_path))
            metrics.append(metric)

        kwargs = {
            "steps": steps,
            "models": models,
            "metrics": metrics
        }

        new_pipeline = cls(**kwargs)
        new_pipeline._best_model_name = config["_best_model_name"]
        return new_pipeline

    @classmethod
    def load(cls, path: str) -> 'Pipeline':
        """
        Load the pipeline

        Parameters
        ----------
        path: str
            path to load the pipeline

        Returns
        -------
        pipeline: Pipeline
            loaded pipeline
        """

        if os.path.exists(path):
            if os.path.isdir(path):
                return cls._load(path)
            else:
                raise ValueError("The path must be a directory")
        else:
            raise ValueError("The path must exist")



from copy import copy

import torch
from torch import nn, relu


class BaselineModel(nn.Module):
    def __init__(self, input_size_proteins, input_size_compounds, hidden_layers_proteins,
                 hidden_layers_compounds, hidden_layers_interaction):
        super().__init__()
        self.hidden_layers_proteins = hidden_layers_proteins
        self.hidden_layers_compounds = hidden_layers_compounds
        self.hidden_layers_interaction = hidden_layers_interaction
        layers_neurons = input_size_proteins * 2
        linear_input = copy(layers_neurons)
        self.dense_proteins1 = nn.Linear(input_size_proteins, layers_neurons)
        for i, hidden_size in enumerate(hidden_layers_proteins):
            self.add_module(
                'fc_proteins{}'.format(i),
                nn.Linear(linear_input, hidden_size)
            )
            linear_input = copy(hidden_size)

        proteins_final_layer = copy(linear_input)

        layers_neurons = input_size_compounds * 2
        self.dense_compounds1 = nn.Linear(input_size_compounds, layers_neurons)
        linear_input = copy(layers_neurons)
        for i, hidden_size in enumerate(hidden_layers_compounds):
            self.add_module(
                'fc_compounds{}'.format(i),
                nn.Linear(linear_input, hidden_size)
            )
            linear_input = copy(hidden_size)
        compounds_final_layer = copy(linear_input)

        self.dense_interaction_layer1 = nn.Linear(proteins_final_layer + compounds_final_layer, layers_neurons)
        linear_input = copy(layers_neurons)
        for i, hidden_size in enumerate(hidden_layers_interaction):
            self.add_module(
                'fc_interaction{}'.format(i),
                nn.Linear(linear_input, hidden_size)
            )
            linear_input = copy(hidden_size)

        self.final_layer = nn.Linear(linear_input, 1)

    def forward(self, x):
        x_proteins = x[0]
        x_proteins = relu(self.dense_proteins1(x_proteins))
        for i, layer in enumerate(self.hidden_layers_proteins):
            x_proteins = relu(getattr(self, 'fc_proteins{}'.format(i))(x_proteins))

        x_compounds = x[1]
        x_compounds = relu(self.dense_compounds1(x_compounds))
        for i, layer in enumerate(self.hidden_layers_compounds):
            x_compounds = relu(getattr(self, 'fc_compounds{}'.format(i))(x_compounds))

        x_interaction = torch.cat([x_proteins, x_compounds], dim=1)
        x_interaction = relu(self.dense_interaction_layer1(x_interaction))
        for i, layer in enumerate(self.hidden_layers_interaction):
            x_interaction = relu(getattr(self, 'fc_interaction{}'.format(i))(x_interaction))
        y = torch.sigmoid(self.final_layer(x_interaction))
        return y


import time
import warnings
from typing import Any, Iterable, List, Union, Dict

import numpy as np
import pandas as pd
from cached_property import cached_property
from pandas import Series

from plants_sm.data_structures.dataset.dataset import Dataset
from plants_sm.io import write_csv
from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.pickle import write_pickle
from plants_sm.mixins.mixins import CSVMixin, ExcelMixin
from plants_sm.data_structures.dataset._utils import process_slices

PLACEHOLDER_FIELD = 'place_holder'


class SingleInputDataset(Dataset, CSVMixin, ExcelMixin):
    _features: Dict[str, Dict[str, np.ndarray]]
    _features_names: List[str]
    _dataframe: Any
    _labels_names: Union[List[str], None] = None
    _instances: Dict[str, dict]
    _features_fields: Dict[str, Union[str, List[Union[str, int]], slice]]
    _identifiers: Union[List[Union[str, int]], None] = None
    _automatically_generated_identifiers: bool = False

    def __init__(self, dataframe: Any = None, representation_field: Union[str, None] = None,
                 features_fields: Union[str, List[Union[str, int]], slice, None] = None,
                 labels_field: Union[str, List[Union[str, int]], slice, None] = None,
                 instances_ids_field: Union[str, None] = None,
                 batch_size: Union[int, None] = None):
        """
        Constructor
        Parameters
        ----------
        dataframe: Any
            dataframe to be consumed by the class and defined as class property
        representation_field: str | List[str | int] (optional)
            representation column field (to be processed)
        features_fields: str | List[str | int] | slice (optional)
            features column field
        labels_field: str | List[str | int] (optional)
            labels column field
        instances_ids_field: str | List[str | int] (optional)
            instances column field
        batch_size: int (optional)
            batch size
        """

        # the features fields is a list of fields that are used to extract the features
        # however, they can be both strings or integers, so we need to check the type of the field
        # and convert it to a list of strings
        super().__init__(batch_size=batch_size)
        if dataframe is not None:
            if not isinstance(features_fields, List) and not isinstance(features_fields, slice) and \
                    features_fields is not None:

                self._features_fields = {PLACEHOLDER_FIELD: [features_fields]}

            elif isinstance(features_fields, slice):

                indexes_list = process_slices(dataframe.columns, features_fields)
                features_fields = [dataframe.columns[i] for i in indexes_list]

                self._features_fields = {PLACEHOLDER_FIELD: features_fields}

            elif features_fields is not None:
                self._features_fields = {PLACEHOLDER_FIELD: features_fields}
            else:
                self._features_fields = {}

            # the instance ids field is defined here, however, if it is None, eventually is derived from the dataframe
            # setter
            self.instances_ids_field = instances_ids_field

            self.representation_field = representation_field

            # the dataframe setter will derive the instance ids field if it is None
            # and also will try to set the features names
            self.dataframe = dataframe
            self._identifiers = self.dataframe[self.instances_ids_field].values
            self.dataframe.index = self.identifiers
            if labels_field is not None:

                if isinstance(labels_field, slice):
                    indexes_list = process_slices(self._dataframe.columns, labels_field)
                    self._labels_names = [self._dataframe.columns[i] for i in indexes_list]

                elif isinstance(labels_field, list):

                    if isinstance(labels_field[0], int):
                        self._labels_names = [self._dataframe.columns[i] for i in
                                              labels_field]

                else:
                    self._labels_names = [labels_field]

                self._labels = None
                # self._labels = self.dataframe.loc[:, self._labels_names].T.to_dict('list')
            else:
                self._labels = None

            if self._features_fields:

                features = self.dataframe.loc[:, self._features_fields[PLACEHOLDER_FIELD]]

                self._features = {PLACEHOLDER_FIELD: {k: features.iloc[i, :].values
                                                      for i, k in enumerate(self._identifiers)}}


            else:
                self._features = {}

            # set the index of the dataframe to the instances ids field
            # self._dataframe.set_index(self.instances_ids_field, inplace=True)

            self.dataframe.drop(self.representation_field, axis=1, inplace=True)
            if self.batch_size is not None:
                while next(self):
                    pass

                self.next_batch()

        # in the case that the dataframe is None and the features field is not None, the features names will be set
    
    def __len__(self):
        return len(self.instances[PLACEHOLDER_FIELD])

    def __getitem__(self, idx):
        return self.instances[PLACEHOLDER_FIELD][idx], self.dataframe.loc[idx, self._labels_names].values

    @classmethod
    def from_csv(cls, file_path: FilePathOrBuffer, representation_field: Union[str, None] = None,
                 features_fields: Union[str, List[Union[str, int]], slice, None] = None,
                 labels_field: Union[str, List[Union[str, int]], slice, None] = None,
                 instances_ids_field: Union[str, None] = None,
                 batch_size: Union[None, int] = None,
                 **kwargs) -> 'SingleInputDataset':
        """
        Method to create a dataset from a csv file.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the csv file
        representation_field: str | List[str | int] (optional)
            representation column field (to be processed)
        features_fields: str | List[str | int] | slice (optional)
            features column field
        labels_field: str | List[str | int] (optional)
            labels column field
        instances_ids_field:
            instances column field
        batch_size: int (optional)
            batch size to be used for the dataset
        kwargs:
            additional arguments for the pandas read_csv method

        Returns
        -------
        dataset: SingleInputDataset
            dataset created from the csv file
        """

        instance = cls()
        dataframe = instance._from_csv(file_path, batch_size, **kwargs)
        dataset = SingleInputDataset(dataframe, representation_field,
                                     features_fields, labels_field,
                                     instances_ids_field,
                                     batch_size=batch_size)
        return dataset

    def to_csv(self, file_path: FilePathOrBuffer, **kwargs):
        """
        Method to write the dataset to a csv file.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the csv file
        kwargs:
            additional arguments for the pandas to_csv method

        Returns
        -------
        success: bool
            True if the file was written successfully, False otherwise
        """

        new_dataframe = self.dataframe.copy()
        data = list(self.instances[PLACEHOLDER_FIELD].items())
        data = pd.DataFrame(data, columns=[self.instances_ids_field, self.representation_field])

        new_dataframe = pd.merge(new_dataframe, data, on=self.instances_ids_field, how='left')

        if self.features:
            write_pkl = False
            features = list(self.features[PLACEHOLDER_FIELD].values())[0]
            if features.ndim > 1:
                warnings.warn(f"The features are not 2D, writing to pickle file")
                write_pkl = True

            if not write_pkl:
                # convert a dictionary into a list of tuples
                data = [(k, *v.tolist()) for k, v in self.features[PLACEHOLDER_FIELD].items()]
                data = pd.DataFrame(data, columns=[self.instances_ids_field] +
                                                  self.features_fields[PLACEHOLDER_FIELD])

                new_dataframe = pd.merge(new_dataframe, data, on=self.instances_ids_field, how='left')

            else:
                write_pickle(file_path.replace("csv", "pkl"), self.features)

        write_csv(file_path, new_dataframe, **kwargs)

    @classmethod
    def from_excel(cls, file_path: FilePathOrBuffer, representation_field: Union[str, None] = None,
                   features_fields: Union[str, List[Union[str, int]], slice, None] = None,
                   labels_field: Union[str, List[Union[str, int]], slice, None] = None,
                   instances_ids_field: Union[str, None] = None,
                   batch_size: Union[int, None] = None, **kwargs) \
            -> 'SingleInputDataset':
        """
        Method to create a dataset from an excel file.

        Parameters
        ----------
        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the csv file
        representation_field: str | List[str | int] (optional)
            representation column field (to be processed)
        features_fields: str | List[str | int] | slice (optional)
            features column field
        labels_field: str | List[str | int] (optional)
            labels column field
        instances_ids_field:
            instances column field
        batch_size: int (optional)
            batch size to be used for the dataset
        kwargs:
            additional arguments for the pandas read_excel method

        Returns
        -------
        dataset: SingleInputDataset
            dataset created from the excel file
        """

        instance = cls()
        dataframe = instance._from_excel(file_path, batch_size=batch_size, **kwargs)
        dataset = SingleInputDataset(dataframe, representation_field,
                                     features_fields, labels_field,
                                     instances_ids_field,
                                     batch_size=batch_size)
        return dataset

    @property
    def identifiers(self) -> List[Union[str, int]]:
        """
        Property for identifiers. It should return the identifiers of the dataset.
        -------
        list of the identifiers: List[Union[str, int]]
        """
        return self._identifiers

    @property
    def features(self) -> Dict[str, Dict[str, np.ndarray]]:
        """
        Property for features. It should return the features of the dataset.
        """
        if self._features is None:
            raise ValueError('Features are not defined')
        return self._features

    @features.setter
    def features(self, value: Dict[str, Dict[str, np.ndarray]]):
        """
        Setter for features.
        Parameters
        ----------
        value : Dict[str, np.ndarray]
            dictionary of the features
        Returns
        -------
        """
        self._clear_cached_properties()
        self._features = value

    def add_features(self, instance_type: str, features: Dict[str, np.ndarray]):
        """
        Add features to the dataset.
        Parameters
        ----------
        instance_type : str
            instance type
        features : Dict[str, np.ndarray]
            dictionary of the features
        Returns
        -------
        """
        if self._features is None:
            self._features = {}
        self._features[instance_type] = features
        self.__dict__.pop('X', None)

    @property
    def features_fields(self):
        return self._features_fields

    @features_fields.setter
    def features_fields(self, value: Dict[str, Union[str, List[Union[str, int]], slice]]):
        """
        Setter for features fields.
        Parameters
        ----------
        value : Union[str, List[Union[str, int]], slice]
            features fields
        Returns
        -------
        """
        self._features_fields = value

    @property
    def labels(self) -> Dict[str, Any]:
        """
        This property will contain the labels for supervised learning.
        Returns
        -------
        Labels for training and prediction. ALIASES: y vector with labels for classification and regression.
        """
        if self._labels_names is not None:
            return self.dataframe.loc[:, self._labels_names].T.to_dict('list')

    @cached_property
    def X(self) -> np.ndarray:
        """
        Property for X. It should return the features of the dataset.
        """
        if self.features == {}:
            raise ValueError('Features are not defined')
        features = [self.features["place_holder"][sequence_id] for sequence_id in
                    self.dataframe[self.instances_ids_field]]
        try:
            return np.array(features, dtype=np.float32)
        except ValueError:
            warnings.warn("The features are not 2D and different dimensions in one of the shapes, "
                          "returning a list of arrays")
            return np.array(features)

    @cached_property
    def y(self) -> np.ndarray:
        """
        Alias for the labels property.
        """
        if not self._labels_names:
            raise ValueError('Labels are not defined')
        # return np.array(list(self.labels.values()))
        return self.dataframe.loc[:, self._labels_names].to_numpy()

    @property
    def instances(self) -> Dict[str, dict]:
        """
        This property will contain the instances of the dataset.
        Returns
        -------
        Array with the instances.
        """
        return self._instances

    def get_instances(self, instance_type: str = PLACEHOLDER_FIELD):
        return self.instances[instance_type]

    @property
    def dataframe(self) -> Any:
        """
        Property of all datasets: they should have an associated dataframe.
        Returns
        -------
        dataframe : Any
            dataframe with the required data
        """
        return self._dataframe

    def set_dataframe(self, value: Any):
        """
        Just a private method to verify the true type of the dataframe according to the type of dataset.

        Parameters
        ----------
        value: Any
            dataframe to be set, it can be in pd.DataFrame format, but can also be a List or Dictionary 
            (it can be specific for each data type)
        Returns
        -------
        """
        if value is not None:
            go = isinstance(value, Iterable) and not isinstance(value, pd.DataFrame) and self.batch_size is not None
            if go:
                # accounting on generators for batch reading
                self._dataframe_generator = value
                value = next(self._dataframe_generator)

            self._set_dataframe(value)
            if self.instances_ids_field is None:
                self._set_instances_ids_field()
            else:
                identifiers = self._dataframe.loc[:, self.instances_ids_field].values
                instances = self.dataframe.loc[:, self.representation_field].values
                self._instances = {PLACEHOLDER_FIELD: dict(zip(identifiers, instances))}
                # self.dataframe.drop(self.representation_field, axis=1, inplace=True)

    @dataframe.setter
    def dataframe(self, value: Any):
        """
        Setter of the property. It verified the type of the value inputted.
        Parameters
        ----------
        value: Any
            dataframe to be set, it can be in pd.DataFrame format, but can also be a List or Dictionary 
            (it can be specific for each data type)
        Returns
        -------
        """
        self.set_dataframe(value)

    def _set_instances_ids_field(self):
        """
        Private method to set the instances ids field if it is not defined.
        """
        if self.instances_ids_field is None:
            self.instances_ids_field = "identifier"
            identifiers_series = Series(list(map(str, range(self.dataframe.shape[0]))), name="identifier", dtype=str)

            if self._features_fields:
                if isinstance(self._features_fields[PLACEHOLDER_FIELD], slice):
                    features_fields_slice = self._features_fields[PLACEHOLDER_FIELD]

                    indexes_list = process_slices(self._dataframe.columns, features_fields_slice)

                    self._features_fields = {PLACEHOLDER_FIELD: [self._dataframe.columns[i] for i in indexes_list]}

                elif isinstance(self._features_fields[PLACEHOLDER_FIELD][0], int):
                    self._features_fields[PLACEHOLDER_FIELD] = [self._dataframe.columns[i] for i in
                                                                self._features_fields[PLACEHOLDER_FIELD]]

            self._automatically_generated_identifiers = True

            self._dataframe = pd.concat((identifiers_series, self._dataframe), axis=1)
            self._dataframe["identifier"] = self._dataframe["identifier"]
            self._identifiers = self._dataframe["identifier"].values

            instances = self.dataframe.loc[:, self.representation_field].values
            self._instances = {PLACEHOLDER_FIELD: dict(zip(identifiers_series.values, instances))}
            # self.dataframe.drop(self.representation_field, axis=1, inplace=True)
        else:
            self._identifiers = self._dataframe[self.instances_ids_field].values

    def select(self, ids: Union[List[str], List[int]], instance_type: str = PLACEHOLDER_FIELD):
        """
        Select a subset of the dataset based on the identifiers.

        Parameters
        ----------
        ids : Union[List[str], List[int]]
            list of identifiers to be selected
        instance_type : str
            type of the instances to be selected
        """

        if self.instances_ids_field is None:
            raise ValueError("Instances ids field is not defined")

        self._dataframe = self._dataframe[self._dataframe[self.instances_ids_field].isin(ids)]
        self._identifiers = self._dataframe[self.instances_ids_field].values

        for instance_type in self._instances:
            self._instances[instance_type] = {k: v for k, v in self._instances[instance_type].items() if k in ids}

            if self._features:
                self._features[instance_type] = {k: v for k, v in self._features[instance_type].items() if k in ids}

        self._clear_cached_properties()

    def merge(self, dataset: 'SingleInputDataset'):

        if self._automatically_generated_identifiers and dataset._automatically_generated_identifiers:
            dataset._identifiers = [f"{i}_" for i in dataset._identifiers]
            # change the identifiers in the dataset instances and features
            dataset._instances[PLACEHOLDER_FIELD] = {k + "_": v for k, v in dataset._instances[PLACEHOLDER_FIELD].items()}

            if self._features:
                dataset._features[PLACEHOLDER_FIELD] = {k + "_": v for k, v in dataset._features[PLACEHOLDER_FIELD].items()}

            dataset._dataframe[self.instances_ids_field] = (
                dataset._dataframe[self.instances_ids_field].apply(lambda x: x + "_"))

        # elif set(self._identifiers).intersection(set(dataset._identifiers)):
        #     raise ValueError("The datasets have common identifiers")

        self._dataframe = pd.concat((self._dataframe, dataset.dataframe), axis=0)

        for instance_type in self._instances:
            self._instances[instance_type].update(dataset._instances[instance_type])

            if self._features:
                self._features[instance_type].update(dataset._features[instance_type])

        self._clear_cached_properties()

        return self



import os
import tempfile
from typing import List, Any
import h5py

from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.reader import Reader
from plants_sm.io.writer import Writer


class H5Reader(Reader):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path
        """
        super().__init__(filepath_or_buffer, mode="r", **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the HDF5 reader can read.

        Returns
        -------
        file_types : List[str]
            object that returns the file types that the HDF5 reader can read.
        """
        return ['h5', 'hdf5']

    def read(self) -> Any:
        """
        Method to read the data from file(s) and return an object.

        Returns
        -------
        object
            The object that was saved in the HDF5 file.
        """
        result_dict = {}
        with h5py.File(self.buffer.name, "r") as f:
            # read object from the HDF5 file
            for key in f.keys():
                result_dict[key] = f[key][:]

        if len(result_dict) == 1:
            return list(result_dict.values())[0]

        return result_dict


class H5Writer(Writer):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path
        """
        super().__init__(filepath_or_buffer, mode="w", **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the HDF5 writer can read.

        Returns
        -------
        file_types : List[str]
            object that returns the file types that the HDF5 writer can read.
        """
        return ['h5', 'hdf5']

    def write(self, object_to_be_written: Any) -> bool:
        """
        Method to write the object into a HDF5 file.

        Parameters
        ----------
        object_to_be_written: Any
            The object to be written.

        Returns
        -------
        bool
            True if the file was written successfully, False otherwise.
        """
        with h5py.File(self.buffer.name, "w") as f:
            # write object to the HDF5 file
            if isinstance(object_to_be_written, dict):
                for key, value in object_to_be_written.items():
                    f.create_dataset(key, data=value)
            else:
                f.create_dataset('data', data=object_to_be_written)

        return True


def read_h5(path: str) -> Any:
    """
    Reads an HDF5 file and returns the object.

    Parameters
    ----------
    path: str
        The path to the HDF5 file.

    Returns
    -------
    Any
        The object that was saved in the HDF5 file.
    """
    return H5Reader(path).read()


def write_h5(object_to_be_written: Any, path: str) -> bool:
    """
    Writes an object into an HDF5 file.

    Parameters
    ----------
    object_to_be_written: Any
        The object to be written.
    path: str
        The path to the HDF5 file.

    Returns
    -------
    bool
        True if the file was written successfully, False otherwise.
    """
    return H5Writer(path).write(object_to_be_written)



import os.path
import re
from abc import abstractmethod
from typing import Any, Union, Iterable

import pandas as pd

from plants_sm.io import read_csv, write_csv
from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.excel import write_excel, read_excel
from plants_sm.io.pickle import write_pickle, read_pickle


class DictMixin:
    def to_dict(self):
        return self._traverse_dict(self.__dict__)

    def _traverse_dict(self, attributes):
        result = {}
        for key, value in attributes.items():
            transverse_value = self._traverse(key, value)
            if transverse_value is not None:
                result[key] = transverse_value

        return result

    def _traverse(self, key, value):
        if not re.match("^_", key):
            if isinstance(value, DictMixin):
                return value.to_dict()
            elif isinstance(value, dict):
                return self._traverse_dict(value)
            elif isinstance(value, list):
                return [self._traverse(key, v) for v in value]
            elif hasattr(value, '__dict__'):
                return self._traverse_dict(value.__dict__)
            else:
                return value
        else:
            return None


class PickleMixin:
    def to_pickle(self, file_path: FilePathOrBuffer) -> bool:
        """
        Method to export the object to pickle.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the object will be exported.

        Returns
        -------
        bool: True if the operation was successful, False otherwise.
        """
        return write_pickle(file_path, self)

    @classmethod
    def from_pickle(cls, file_path: FilePathOrBuffer) -> 'PickleMixin':
        """
        Method to import the object from pickle.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the object will be imported.

        Returns
        -------
        PickleMixin: object imported from pickle.
        """
        return read_pickle(file_path)


class CSVMixin:

    def __init__(self):
        self._dataframe = None

    @property
    @abstractmethod
    def dataframe(self) -> Any:
        """
        Abstract method and property that returns the dataframe.
        Returns
        -------

        """
        pass

    @dataframe.setter
    def dataframe(self, value: Any):
        """
        Setter for dataframe.

        Parameters
        ----------
        value: Any
            value to be set as dataframe
        """
        self._dataframe = value

    def to_csv(self, file_path: FilePathOrBuffer, **kwargs) -> bool:
        """
        Method to export the dataframe to csv.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be exported.

        Returns
        -------
        bool: True if the operation was successful, False otherwise.

        Raises
        ------
        NotImplementedError
            If the dataframe is not a type where the package cannot export it.

        """

        if isinstance(self.dataframe, pd.DataFrame):
            return write_csv(filepath_or_buffer=file_path,
                             df=self.dataframe,
                             **kwargs)
        else:
            raise NotImplementedError("This method is not implemented for this type of object")

    @staticmethod
    def retrieve_a_generator(file_path: str, batch_size: Union[None, int], **kwargs) -> Iterable:
        """
        Method to retrieve a generator from a csv file.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be imported.
        batch_size: Union[None, int]
            size of the batch to be read from the csv file.
        kwargs: dict
            arguments to be passed to the read_csv method.

        Returns
        -------
        Iterable: generator that will return the batches of the dataframe.
        """
        gen = read_csv(file_path, chunksize=batch_size, iterator=True, get_buffer=False, **kwargs)
        for batch in gen:
            yield batch

    @staticmethod
    def _from_csv(file_path: FilePathOrBuffer, batch_size: Union[None, int] = None, **kwargs) -> Any:
        """
        Method to import the dataframe from csv.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be imported.
        batch_size: Union[None, int]
            size of the batch to be read from the csv file.
        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} does not exist.")

        if not batch_size:
            df = read_csv(file_path, **kwargs)
            return df

        else:
            return CSVMixin.retrieve_a_generator(file_path, batch_size, **kwargs)

    @classmethod
    def from_csv(cls, file_path: FilePathOrBuffer, **kwargs) -> 'CSVMixin':
        """
        Method to import the dataframe from csv.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be imported.

        Returns
        -------
        CSVMixin: object of the class that inherits from CSVMixin

        """


class ExcelMixin:

    def __init__(self):
        self._dataframe = None

    @property
    @abstractmethod
    def dataframe(self) -> Any:
        """
        Abstract method and property that returns the dataframe.
        Returns
        -------

        """
        pass

    @dataframe.setter
    def dataframe(self, value: Any):
        """
        Setter for dataframe.

        Parameters
        ----------
        value: Any
            value to be set as dataframe
        """
        self._dataframe = value

    def to_excel(self, file_path: FilePathOrBuffer, **kwargs) -> bool:
        """
        Method to export the dataframe to excel.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be exported.

        Returns
        -------
        bool: True if the file was created, False otherwise.

        """

        if isinstance(self.dataframe, pd.DataFrame):
            return write_excel(file_path, self.dataframe, **kwargs)
        else:
            raise NotImplementedError("This method is not implemented for this type of object")

    @staticmethod
    def retrieve_a_generator(file_path, batch_size, **kwargs) -> Iterable:
        """
        Method to retrieve a generator from a csv file.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be imported.
        batch_size: Union[None, int]
            size of the batch to be read from the csv file.
        kwargs: dict
            arguments to be passed to the read_csv method.

        Returns
        -------
        Iterable: generator that will return the batches of the dataframe.
        """
        gen = read_excel(file_path, chunksize=batch_size, iterator=True, get_buffer=False, **kwargs)
        for batch in gen:
            yield batch

    @staticmethod
    def _from_excel(file_path: FilePathOrBuffer, batch_size=Union[None, int], **kwargs) -> Any:
        """
        Method to import the dataframe from excel.
        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be imported.

        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} does not exist.")

        if not batch_size:

            df = read_excel(file_path, **kwargs)
            return df

        else:
            return ExcelMixin.retrieve_a_generator(file_path, batch_size, **kwargs)

    @classmethod
    def from_excel(cls, file_path: FilePathOrBuffer, **kwargs) -> 'ExcelMixin':
        """
        Method to import the dataframe from excel.

        Parameters
        ----------
        file_path: FilePathOrBuffer
            path to the file where the dataframe will be imported.

        Returns
        -------
        ExcelMixin: object of the class that inherits from ExcelMixin

        """



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.cluster import KMeans

from plants_sm.unsupervised.model import UnsupervisedModel


class SklearnKMeans(UnsupervisedModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.kmeans = KMeans(**self.kwargs)

    def _fit(self, x):
        self.kmeans.fit(x)
        self.fitted = True  # add decorator
        return self

    def _transform(self, x) -> pd.DataFrame:
        if self.fitted:
            return self.kmeans.transform(x)
        else:
            raise ValueError("The model is not fitted")

    def _predict(self, x):
        if self.fitted:
            return self.kmeans.predict(x)
        else:
            raise ValueError("The model is not fitted")

    def fit_predict_generate_scatter_plot(self, x):
        labels = self.kmeans.fit_predict(x)
        u_labels = np.unique(labels)
        plt.subplots(1, figsize=(25, 15))
        # plotting the results:
        centroids = self.kmeans.cluster_centers_
        for i in u_labels:
            plt.scatter(x[labels == i, 0], x[labels == i, 1], label=i, s=40)
        plt.scatter(centroids[:, 0], centroids[:, 1], s=80, color='k')
        plt.title('KMeans')
        plt.legend()

    def generate_distortion_graph(self, x):
        distortions = []
        for i in range(1, 21):
            km = KMeans(
                **self.kwargs
            )
            km.fit(x)
            distortions.append(km.inertia_)

        # plot
        plt.plot(range(1, 21), distortions, marker='o')
        plt.xlabel('Number of clusters')
        plt.ylabel('Distortion')
        plt.title('Elbow Method')
        plt.show()



import os
from typing import Union

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.io.pickle import read_pickle
from plants_sm.models._utils import _convert_proba_to_unified_form, write_model_parameters_to_pickle, _batch_generator
from plants_sm.models.constants import REGRESSION, QUANTILE, FileConstants
from plants_sm.models.model import Model

import tensorflow as tf


class TensorflowModel(Model):

    def __init__(self, model: tf.keras.Model, epochs: int, batch_size: int, problem_type: str,
                 callbacks: list = None):
        """
        Constructor for TensorflowModel class.

        Parameters
        ----------
        model: tf.keras.Model
            Tensorflow model to be used for training and prediction.
        epochs: int
            Number of epochs to train the model.
        batch_size: int
            Batch size to be used for training.
        problem_type: str
            Type of problem to be solved. Can be 'binary', 'multiclass','regression', 'softclass', 'quantile'.
        callbacks: list
            List of callbacks to be used during training.
        """
        self.model = model
        self.epochs = epochs
        self.batch_size = batch_size
        self.callbacks = callbacks
        self.problem_type = problem_type
        self._history = None

    def _preprocess_data(self, dataset: Dataset, **kwargs):
        """
        Preprocess the data.

        Parameters
        ----------
        dataset: Dataset
            Dataset to be used for training.
        kwargs: dict
            Keyword arguments to be used for preprocessing.

        """
        pass

    def _fit_data(self, train_dataset: Dataset, validation_dataset: Dataset) -> None:
        """
        Fit the model to the data.

        Parameters
        ----------
        train_dataset: Dataset
            Dataset to be used for training.
        validation_dataset: Dataset
            Dataset to be used for validation.
        """
        self._preprocess_data(train_dataset)
        self._preprocess_data(validation_dataset)

        if train_dataset.batch_size is None:
            if isinstance(train_dataset.X, dict):
                train_dataset_x = train_dataset.X.values()
            else:
                train_dataset_x = train_dataset.X
            train_dataset_y = train_dataset.y
            train_dataset_tuple = (train_dataset_x, train_dataset_y)
        else:
            train_dataset_tuple = _batch_generator(train_dataset)

        if validation_dataset.batch_size is None:
            if isinstance(validation_dataset.X, dict):
                validation_dataset_x = validation_dataset.X.values()
            else:
                validation_dataset_x = validation_dataset.X
            validation_dataset_y = validation_dataset.y
            validation_dataset_tuple = (validation_dataset_x, validation_dataset_y)
        else:
            validation_dataset_tuple = _batch_generator(validation_dataset)

        if train_dataset.batch_size is None:
            self._history = self.model.fit(train_dataset_tuple[0], train_dataset_tuple[1],
                                           epochs=self.epochs,
                                           batch_size=self.batch_size,
                                           callbacks=self.callbacks,
                                           validation_data=validation_dataset_tuple)

        else:
            self._history = self.model.fit(train_dataset_tuple,
                                           epochs=self.epochs,
                                           batch_size=self.batch_size,
                                           callbacks=self.callbacks,
                                           validation_data=validation_dataset_tuple)

    def _predict_proba(self, dataset: Dataset) -> np.ndarray:
        """
        Predict probabilities for each class.

        Parameters
        ----------
        dataset: Dataset
            Dataset to be used for prediction.

        Returns
        -------
        np.ndarray
            Predicted probabilities for each class.
        """
        if dataset.batch_size is None:
            if isinstance(dataset.X, dict):
                y_pred = self.model.predict(dataset.X.values())
            else:
                y_pred = self.model.predict(dataset.X)
        else:
            y_pred = self.model.predict(_batch_generator(dataset))

        return y_pred

    def _predict(self, dataset: Dataset) -> np.ndarray:
        """
        Predict the class for each sample.

        Parameters
        ----------
        dataset: Dataset
            Dataset to be used for prediction.

        Returns
        -------
        np.ndarray
            Predicted class for each sample.
        """

        if dataset.batch_size is None:
            if isinstance(dataset.X, dict):
                y_pred = self.model.predict(dataset.X.values())
            else:
                y_pred = self.model.predict(dataset.X)

        else:
            y_pred = self.model.predict(_batch_generator(dataset))

        if self.problem_type in [REGRESSION, QUANTILE]:
            return y_pred
        else:
            return _convert_proba_to_unified_form(self.problem_type, np.array(y_pred))

    def _save(self, path: str) -> None:
        """
        Save the model to the specified path.

        Parameters
        ----------
        path: str
            Path to save the model.
        """
        self.model.save(os.path.join(path, FileConstants.TENSORFLOW_MODEL.value))
        # get the class attributes and save them
        parameters = {'epochs': self.epochs,
                      'batch_size': self.batch_size,
                      'callbacks': self.callbacks,
                      'problem_type': self.problem_type}
        write_model_parameters_to_pickle(parameters, path)

    @classmethod
    def _load(cls, path: str) -> 'TensorflowModel':
        """
        Load the model from the specified path.

        Parameters
        ----------
        path: str
            Path to load the model from.
        """
        model = tf.keras.models.load_model(os.path.join(path, FileConstants.TENSORFLOW_MODEL.value))
        parameters = read_pickle(os.path.join(path, FileConstants.MODEL_PARAMETERS_PKL.value))
        return cls(model, **parameters)

    @property
    def history(self) -> Union[dict, any, None]:
        """
        Get the history of the model.

        Returns
        -------
        dict
            History of the model.
        """
        if self._history is None:
            return None
        else:
            return self._history



import warnings

from cached_property import cached_property
from typing import List, Any, Union, Dict, Iterable

import numpy as np
import pandas as pd

from plants_sm.data_structures.dataset import Dataset
from plants_sm.io import write_csv
from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.pickle import write_pickle
from plants_sm.mixins.mixins import CSVMixin, ExcelMixin
from plants_sm.data_structures.dataset._utils import process_slices


class MultiInputDataset(Dataset, CSVMixin, ExcelMixin):
    _features: Dict[str, Dict[str, np.ndarray]]
    _features_names: List[str]
    _dataframe: pd.DataFrame
    _representation_field: str
    _labels_names: List[str] = None
    _features_fields: Dict[str, Union[str, int]] = {}
    _instances: Dict[str, Dict[str, Any]]
    _identifiers: List[Union[str, int]] = None

    def __init__(self, dataframe: Any = None, representation_field: Dict[str, Union[str, int]] = None,
                 labels_field: Union[str, List[Union[str, int]]] = None,
                 instances_ids_field: Dict[str, Union[str, int]] = None,
                 batch_size: int = None):
        """
        Constructor
        Parameters
        ----------
        dataframe: Any
            dataframe to be consumed by the class and defined as class property
        representation_field: str | List[str | int] (optional)
            representation column field (to be processed)
        labels_field: str | List[str | int] (optional)
            labels column field
        instances_ids_field: str | List[str | int] (optional)
            instances column field
        batch_size: int (optional)
            the size of the batches
        """

        # the features fields is a list of fields that are used to extract the features
        # however, they can be both strings or integers, so we need to check the type of the field
        # and convert it to a list of strings
        super().__init__(batch_size=batch_size)
        if dataframe is not None:
            self._features = {}
            self._labels = None
            self.instances_ids_field = instances_ids_field
            self.representation_field = representation_field

            # the dataframe setter will derive the instance ids field if it is None
            # and also will try to set the features names
            self.dataframe = dataframe

            if labels_field is not None:

                if isinstance(labels_field, slice):
                    indexes_list = process_slices(self._dataframe.columns, labels_field)
                    self._labels_names = [self._dataframe.columns[i] for i in indexes_list]

                elif isinstance(labels_field, list):

                    if isinstance(labels_field[0], int):
                        self._labels_names = [self._dataframe.columns[i] for i in
                                              labels_field]

                else:
                    self._labels_names = [labels_field]

            self.dataframe.drop(list(self.representation_field.values()), axis=1, inplace=True)
            if self.batch_size is not None:
                while next(self):
                    pass
                self.next_batch()

    @property
    def dataframe(self) -> Any:
        return self._dataframe

    @classmethod
    def from_csv(cls, file_path: FilePathOrBuffer, representation_field: Dict[str, Union[str, int]] = None,
                 labels_field: Union[str, List[Union[str, int]]] = None,
                 instances_ids_field: Dict[str, Union[str, int]] = None,
                 batch_size: Union[None, int] = None,
                 **kwargs: object) -> 'MultiInputDataset':

        instance = cls()
        dataframe = instance._from_csv(file_path, batch_size, **kwargs)
        dataset = MultiInputDataset(dataframe, representation_field,
                                    labels_field, instances_ids_field,
                                    batch_size=batch_size)
        return dataset

    @classmethod
    def from_excel(cls, file_path: FilePathOrBuffer,
                   representation_field: Dict[str, Union[str, int]] = None,
                   labels_field: Union[str, List[Union[str, int]]] = None,
                   instances_ids_field: Dict[str, Union[str, int]] = None,
                   batch_size: Union[None, int] = None, **kwargs) -> 'MultiInputDataset':

        instance = cls()
        dataframe = instance._from_excel(file_path, batch_size, **kwargs)
        dataset = MultiInputDataset(dataframe, representation_field,
                                    labels_field, instances_ids_field,
                                    batch_size=batch_size)
        return dataset

    @property
    def instances(self):
        return self._instances

    @property
    def identifiers(self):
        """
        The identifiers of the instances (interaction between the instances)
        """
        return self._identifiers

    @identifiers.setter
    def identifiers(self, value: List[Union[str, int]]):
        """
        Setter of the property. It verified the type of the value inputted.

        Parameters
        ----------
        value: List[Union[str, int]]
            identifiers to be set

        """
        self._identifiers = value

    @property
    def labels(self):
        return self._labels

    @property
    def features_fields(self):
        return self._features_fields

    def get_instances(self, instance_type: str = None):
        if instance_type is None:
            raise ValueError("The instance type must be specified")
        return self._instances[instance_type]

    @dataframe.setter
    def dataframe(self, value: Any):
        """
        Setter of the property. It verified the type of the value inputted.
        Parameters
        ----------
        value: Any
            dataframe to be set, it can be in pd.DataFrame format, but can also be a List or Dictionary
            (it can be specific for each data type)
        Returns
        -------
        """
        self.set_dataframe(value)

    def set_dataframe(self, value: Any):
        """
        Just a private method to verify the true type of the dataframe according to the type of dataset.

        Parameters
        ----------
        value: Any
            dataframe to be set, it can be in pd.DataFrame format, but can also be a List or Dictionary
            (it can be specific for each data type)
        Returns
        -------
        """
        if value is not None:
            go = isinstance(value, Iterable) and not isinstance(value, pd.DataFrame) and self.batch_size is not None
            if go:
                # accounting on generators for batch reading
                self._dataframe_generator = value
                value = next(self._dataframe_generator)

            self._set_dataframe(value)
            self.set_instances_and_ids()

    def set_instances_and_ids(self):
        """
        Set the instances and the identifiers of the instances
        """
        self._instances = {}
        instance_types = []
        for instance_type, instance_field in self.representation_field.items():
            if instance_type not in self.instances_ids_field:
                self.generate_ids(instance_type)
            elif self.instances_ids_field[instance_type] not in self.dataframe.columns:
                self.generate_ids(instance_type)

            identifiers = self.dataframe.loc[:, self.instances_ids_field[instance_type]].values
            if instance_type not in self._instances:
                instances = self.dataframe.loc[:, instance_field].values

                self._instances[instance_type] = dict(zip(identifiers, instances))
                instance_types.append(instance_type)

        if self._identifiers is None:
            _identifiers_ids = []
            for id_type, ids_field in self.instances_ids_field.items():
                if id_type not in instance_types:
                    _identifiers_ids.append(ids_field)

            self._identifiers = self.dataframe.loc[:, _identifiers_ids].values

    def generate_ids(self, instance_type):
        """
        Generates a new identifiers for the instance type
        Parameters
        ----------
        instance_type: str
            the instance type
        Returns
        -------
        int
            the new identifier
        """
        # generate identifiers for unique elements
        unique_elements = np.unique(self.dataframe.loc[:, self.representation_field[instance_type]].values)
        new_ids = np.arange(1, len(unique_elements) + 1)

        self.dataframe.loc[:, f"{instance_type}_identifiers"] = \
            self.dataframe.loc[:, self.representation_field[instance_type]].map(dict(zip(unique_elements, new_ids)))

        self.instances_ids_field[instance_type] = f"{instance_type}_identifiers"

    @property
    def features(self) -> Dict[str, Dict[str, np.ndarray]]:
        """
        Property for features. It should return the features of the dataset.
        """
        return self._features

    @features.setter
    def features(self, value: Dict[str, Dict[str, np.ndarray]]):
        """
        Setter for features.
        Parameters
        ----------
        value : Dict[str, np.ndarray]
            dictionary of the features
        Returns
        -------
        """
        self._features = value
        self._clear_cached_properties()
        self.__dict__.pop('X', None)

    def add_features(self, instance_type: str, features: Dict[str, np.ndarray]):
        """
        Adds features to the dataset
        Parameters
        ----------
        instance_type: str
            the instance type
        features: Dict[str, np.ndarray]
            the features
        Returns
        -------
        """
        if instance_type in self._features:
            self._features[instance_type].update(features)
        else:
            self._features[instance_type] = features

        self.__dict__.pop('X', None)

    def to_csv(self, file_path: FilePathOrBuffer, **kwargs):
        """
        Saves the dataset to a csv file
        Parameters
        ----------
        file_path: FilePathOrBuffer
            the path of the file
        kwargs: dict
            the arguments for the pandas.to_csv method
        Returns
        -------
        bool
            True if the operation is successful
        """
        new_dataframe = self.dataframe.copy()
        for instance_type in self.instances.keys():
            data = list(self.instances[instance_type].items())
            data = pd.DataFrame(data, columns=[self.instances_ids_field[instance_type],
                                               self.representation_field[instance_type]])

            new_dataframe = pd.merge(new_dataframe, data, on=self.instances_ids_field[instance_type], how='left')

        try:
            instances_types = self.X.keys()
        except ValueError:
            instances_types = None

        if self._features_fields is not None and instances_types is not None:
            write_pkl = False
            for instance_type in instances_types:
                instance_features = self.X[instance_type]

                if instance_features.ndim > 2:
                    warnings.warn(f"The features of the instance {instance_type} are not 2D, writing to pickle file")
                    write_pkl = True
                    break

                new_dataframe = pd.concat([new_dataframe, pd.DataFrame(instance_features,
                                                                       columns=self.features_fields[instance_type])],
                                          axis=1)

            if write_pkl:
                write_pickle(file_path.replace("csv", "pkl"), self.features)

        write_csv(file_path, new_dataframe, **kwargs)

    @cached_property
    def X(self):
        """
        Property for X. It should return the features of the dataset.

        Returns
        -------
        Dict[str, np.ndarray]
            the features, the key is the instance type and the value is the features
        """

        res = {}
        feature_keys = self.features.keys()
        for instance_type, instance in self.instances.items():  # probably 2 or a little more instances
            res[instance_type] = []

            if instance_type in feature_keys:
                instance_ids = self.dataframe.loc[:, self.instances_ids_field[instance_type]]
                for instance_id in instance_ids:
                    res[instance_type].append(self.features[instance_type][instance_id])

                res[instance_type] = np.array(res[instance_type])
            else:
                raise ValueError(f"Features for {instance_type} are not defined")
        return res

    @property
    def y(self):
        if self._labels_names is not None:
            return self.dataframe.loc[:, self._labels_names].to_numpy()

    def select(self, ids: Union[List[str], List[int]], instance_type: str):
        """
        Selects the instances with the given ids

        Parameters
        ----------
        ids: Union[List[str], List[int]]
            the ids of the instances to select
        instance_type: str
            the instance type
        """
        self._dataframe = self._dataframe[self._dataframe[self.instances_ids_field[instance_type]].isin(ids)]
        instance_types = []
        for instance_type in self._instances:
            instance_ids = self._dataframe.loc[:, self.instances_ids_field[instance_type]].values
            self._instances[instance_type] = {k: v for k, v in self._instances[instance_type].items()
                                              if k in instance_ids}

            if self._features:
                self._features[instance_type] = {k: v for k, v in self._features[instance_type].items()
                                                    if k in instance_ids}

            instance_types.append(instance_type)

        _identifiers_ids = []
        for id_type, ids_field in self.instances_ids_field.items():
            if id_type not in instance_types:
                _identifiers_ids.append(ids_field)

        self._identifiers = self.dataframe.loc[:, _identifiers_ids].values
        self._clear_cached_properties()

    def merge(self, dataset: 'MultiInputDataset'):
        """
        Merges the dataset with another dataset

        Parameters
        ----------
        dataset: MultiInputDataset
            the dataset to merge
        """
        self._dataframe = pd.concat([self._dataframe, dataset.dataframe], ignore_index=True)
        self._instances.update(dataset.instances)
        for instance_type in self._instances:
            self._instances[instance_type].update(dataset._instances[instance_type])

            if self._features:
                self._features[instance_type].update(dataset._features[instance_type])
        self._identifiers = self.dataframe.loc[:, self.instances_ids_field.values()].values
        self._clear_cached_properties()

        return self



from torch import nn

class DNN(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size, batch_norm=False, last_sigmoid=True, dropout=None):
        super(DNN, self).__init__()

        self.batch_norm = batch_norm

        self.hidden_sizes = hidden_sizes
        
        if len(hidden_sizes) != 0:
            self.fc_initial = nn.Linear(input_size, hidden_sizes[0])
            self.batch_norm_initial = nn.BatchNorm1d(hidden_sizes[0])
            self.relu_initial = nn.ReLU()
            self.last_sigmoid = last_sigmoid
            if dropout is not None:
                self.dropout = nn.Dropout(dropout)
            else:
                self.dropout = None

            for i in range(1, len(hidden_sizes)):
                setattr(self, f"fc{i}", nn.Linear(hidden_sizes[i-1], hidden_sizes[i]))
                setattr(self, f"relu{i}", nn.ReLU())
                setattr(self, f"batch_norm_layer{i}", nn.BatchNorm1d(hidden_sizes[i]))
        

            self.fc_final = nn.Linear(hidden_sizes[-1], output_size)
        else:
            # self.fc_initial = nn.Linear(input_size, output_size)
            # self.batch_norm_initial = nn.BatchNorm1d(output_size)
            self.relu_initial = nn.ReLU()
            self.last_sigmoid = last_sigmoid
            if dropout is not None:
                self.dropout = nn.Dropout(dropout)
            else:
                self.dropout = None

            self.fc_final = nn.Linear(input_size, output_size)

        self.final_relu = nn.ReLU()
        self.final_batch_norm = nn.BatchNorm1d(output_size)

    def forward(self, x):

        if type(x) == list:
            x = x[0]

        if len(self.hidden_sizes) != 0:
            out = self.fc_initial(x)
            if self.batch_norm and x.shape[0] > 1:
                out = self.batch_norm_initial(out)
            out = self.relu_initial(out)
            if self.dropout is not None:
                out = self.dropout(out)
            for i in range(1, len(self.hidden_sizes)):
                out = getattr(self, f"fc{i}")(out)
                if self.batch_norm and x.shape[0] > 1:
                    out = getattr(self, f"batch_norm_layer{i}")(out)
                out = getattr(self, f"relu{i}")(out)
                if self.dropout is not None:
                    out = self.dropout(out)
            
            out = self.fc_final(out)
        else:
            out = self.fc_final(x)
        
        if self.last_sigmoid:
            out = nn.Sigmoid()(out)
        return out



import io
import pickle
from typing import List, Any

import torch

from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.reader import Reader
from plants_sm.io.writer import Writer

import tracemalloc



class PickleReader(Reader):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path
        """
        super().__init__(filepath_or_buffer, mode="rb", **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the pickle reader can read.

        Returns
        -------
        file_types : List[str]
            object that returns the file types that the pickle reader can read.
        """
        return ['pkl']

    def read(self) -> Any:
        """
        Method to read the data from file(s) and returns a pandas pd.DataFrame.

        Returns
        -------

        """

        class CPU_Unpickler(pickle.Unpickler):
            def find_class(self, module, name):
                if module == 'torch.storage' and name == '_load_from_bytes':
                    return lambda b: torch.load(io.BytesIO(b), map_location='cpu')
                else:
                    return super().find_class(module, name)

        contents = CPU_Unpickler(self.buffer).load()
        # pickled_object = pickle.load(self.buffer)
        self.close_buffer()
        return contents


class PickleWriter(Writer):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path
        """
        super().__init__(filepath_or_buffer, mode="wb", **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the pickle writer can read.

        Returns
        -------
        file_types : List[str]
            object that returns the file types that the pickle writer can read.
        """
        return ['pkl']

    def write(self, object_to_be_written: Any) -> bool:
        """
        Method to write the pd.DataFrame into a pickle file.

        Parameters
        ----------
        object_to_be_written: Any
            object to be written

        Returns
        -------
        bool
            True if the file was written successfully, False otherwise.
        """
        pickle.dump(object_to_be_written, self.buffer, protocol=pickle.HIGHEST_PROTOCOL)
        # pickle.dump(object_to_be_written, self.buffer)
        self.close_buffer()
        return True


def read_pickle(path: str) -> Any:
    """
    Reads a pickle file and returns the object.

    Parameters
    ----------
    path: str
        The path to the pickle file.

    Returns
    -------
    Any
        The object that was saved in the pickle file.
    """
    reader = PickleReader(path)
    pickled_object = reader.read()
    return pickled_object


def write_pickle(path: str, object_to_be_written: Any) -> bool:
    """
    Writes an object into a pickle file.

    Parameters
    ----------
    path: str
        The path to the pickle file.
    object_to_be_written: Any
        The object to be written.

    Returns
    -------
    bool
        True if the file was written successfully, False otherwise.
    """
    tracemalloc.start()
    return PickleWriter(path).write(object_to_be_written)


def is_pickable(object_to_be_written: Any):
    """
    Checks if an object is pickable.

    Parameters
    ----------
    object_to_be_written: Any
        The object to be written.

    Returns
    -------

    """
    try:
        pickle.dumps(object_to_be_written)
        return True
    except Exception:
        return False
